import numpy as np
from arthurai.common.constants import OutputType, Stage
from arthurai.common.exceptions import MethodNotApplicableError


def get_pred_and_gt_attrs(arthur_model):
    """Returns the name of the predicted attribute and ground truth attribute for a model.

    :param arthur_model: an ArthurModel object
    """
    if arthur_model.output_type == OutputType.Regression:
        return arthur_model.get_attributes(Stage.PredictedValue)[0].name , arthur_model.get_attributes(Stage.GroundTruth)[0].name
    elif arthur_model.output_type == OutputType.Multiclass:
        return [(attr.name, attr.attribute_link) for attr in arthur_model.get_attributes(Stage.PredictedValue) if attr.is_positive_predicted_attribute][0]
    else:
        raise MethodNotApplicableError()

# coefficients for cubic polyomial with window size of 5
savgol_coefs = np.array([-0.08571429,  0.34285714,  0.48571429,  0.34285714, -0.08571429])

def savgol_filter(signal):
    """ Simple instantiation of Savitzky-Golay filter (cubic).

    :param signal: a numpy 1D array
    """ 
    return np.convolve(signal, savgol_coefs, mode="same")


import json
from dataclasses import asdict
from typing import Union

from dataclasses_json import DataClassJsonMixin


class ArthurBaseJsonDataclass(DataClassJsonMixin):
    @staticmethod
    def clean_nones(d):
        if not isinstance(d, (dict, list)):
            return d
        if isinstance(d, list):
            return list(filter(lambda item: item is not None, map(ArthurBaseJsonDataclass.clean_nones, d)))
        return {k: v for k, v in ((k, ArthurBaseJsonDataclass.clean_nones(v)) for k, v in d.items()) if v is not None }

    def to_json(self, skip_none=True) -> str: # type: ignore
        obj_dict = asdict(self)
        if skip_none:
            obj_dict = self.clean_nones(obj_dict)
        return json.dumps(obj_dict)

    def to_dict(self, skip_none=True) -> dict:
        """Creates a dictionary representation of the attribute object

        :return: Dictionary object of attribute data
        """
        obj_dict = asdict(self)
        if skip_none:
            obj_dict = self.clean_nones(obj_dict)
        return obj_dict


NumberType = Union[int, float]


# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

from typing import List

sys.path.insert(0, os.path.abspath('../../'))
sys.path.insert(0, os.path.abspath('../../arthurai/'))
sys.path.append(os.path.abspath("../sphinx_ext"))


# -- Project information -----------------------------------------------------

project = 'ArthurAI SDK'
copyright = '2022, ArthurAI'
author = 'ArthurAI'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['sphinx.ext.autodoc',
              # note: autodoc has native typehint support now but result isn't as pretty and it mis-orders permissions
              'sphinx_autodoc_typehints',
              'autosummary_permissions']
              # DO NOT UNCOMMENT unless testing! our autosummary_permissions extension will setup regular autosummary
              # 'sphinx.ext.autosummary']

# these are the defaults but to be explicit
autosummary_generate = True
autosummary_imported_members = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns: List[str] = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'furo'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']


# The light_logo and dark_logo properties take a file path relative to the _static folder
html_theme_options = {
    "light_logo": "images/light-mode-logo.svg",
    "dark_logo": "images/dark-mode-logo.svg",
    "dark_css_variables": {
        "color-brand-primary": "#AC37F6",
        "color-brand-content": "#AC37F6",
        "color-problematic": "#AC37F6",
    },
    "light_css_variables": {
        "color-brand-primary": "#7C4DFF",
        "color-brand-content": "#7C4DFF",
        "color-problematic": "#7C4DFF",
    },
    "sidebar_hide_name": True,

}

# Custom CSS file in case we want to add more styling
html_css_files = [
    'css/custom_styles.css'
]



from typing import List, Callable, Dict, Iterable, Any, Optional, Tuple, Union
import logging

import numpy as np
import pandas as pd

from lime.lime_tabular import LimeTabularExplainer
from lime.explanation import Explanation
from lime.lime_text import LimeTextExplainer
from lime.lime_image import LimeImageExplainer
from shap import KernelExplainer
import skimage.segmentation as seg
import re

from arthurai.common.constants import OutputType, InputType, TextDelimiter
from arthurai.common.exceptions import UserValueError, MethodNotApplicableError, ArthurUserError

logger = logging.getLogger(__name__)


class EmptyLimeExplanation:
    """
    Class for representing an all-NaN LIME Explanation. The LIME Explainers return full-featured
    lime.explanation.Explanation objects. We only use these to fetch the feature importance scores for each predicted
    class by calling `as_map()`. This class implements that `as_map()` method, and returns all-NaN feature importance
    scores.
    """
    def __init__(self, class_int_labels: List[int], num_features: int):
        nans: List[Tuple[int, float]] = [(feat_idx, float("nan")) for feat_idx in range(num_features)]
        self.nan_dict: Dict[int, List[Tuple[int, float]]] = {label: nans for label in class_int_labels}

    def as_map(self) -> Dict[int, List[Tuple[int, float]]]:
        """
        :return: A map from each predicted class integer label to a list of NaNs
        """
        return self.nan_dict


class ArthurExplainer:
    """Class for explaining inferences with a variety of libs"""

    SHAP = 'shap'
    LIME = 'lime'

    def __init__(self,
                 model_type: OutputType,
                 model_input_type: InputType,
                 num_predicted_attributes: int,
                 predict_func: Callable,
                 data: Optional[pd.DataFrame],
                 label_mapping: List[Optional[Dict[str, int]]],
                 categorical_features: Optional[List[int]] = None,
                 enable_shap: bool = True,
                 shap_training_samples: int = 20,
                 enable_lime: bool = True,
                 text_delimiter: TextDelimiter = TextDelimiter.NOT_WORD,
                 load_image_func: Optional[Callable] = None):

        """
        :param model_type: Type of model being explained
        :param model_input_type: Type of input data, tabular, nlp, image
        :param predict_func: callable, predict function for model
        :param data: Pandas DataFrame, required for tabular models
        :param label_mapping: A list of optional dictionaries. Each index corresponds to the feature index that will
            use the mapping
        :param categorical_features: A list of indices of the categorical features in the input data
        :param enable_shap: bool, should shap explanations be enabled, default true
        :param shap_training_samples: number of training samples to initialize shap
        :param enable_lime: bool, should Lime explanations be enabled, default true
        :param lime_text_bow: bool, should Lime treat input as a bag of words. Set to false if using n-grams, default true
        :param text_delimiter: TextDelimiter, the pattern used to split raw document into tokens, passed to `re.split`, default NOT_WORD
        """

        self.model_type = model_type
        self.model_input_type = model_input_type
        self.num_predicted_attributes = num_predicted_attributes
        self.input_predict_func = predict_func
        self.label_mapping = label_mapping
        self.categorical_features = categorical_features
        self.shap_training_samples = shap_training_samples
        self.reverse_label_mapping = [
            None if mapping is None else {val: key for key, val in mapping.items()}
            for mapping in self.label_mapping
        ]
        self.text_delimiter = text_delimiter
        self.load_image_func = load_image_func
        self.enable_shap = enable_shap
        self.enable_lime = enable_lime

        # ensure there are no nans in data
        if data is not None:
            length_with_nans = len(data)
            data = data.dropna()
            if len(data) != length_with_nans:
                logger.warning(f"Dropped {length_with_nans - len(data)} rows with NaN values from explainer dataset")

        # encode dataframe, don't save this as a instance variable for size
        # image models don't require dataframe
        if model_input_type != InputType.Image:
            encoded_data = pd.DataFrame([self.encode_labels(fv) for fv in data.to_numpy()])
        else:
            encoded_data = None

        self.tabular_explainer_shap = None
        self.tabular_explainer_lime = None

        # build tabular explainers
        if self.model_input_type == InputType.Tabular:
            self.build_tabular_explainers(enable_lime=self.enable_lime, enable_shap=self.enable_shap, data=encoded_data)
        if self.model_input_type == InputType.NLP:
            if self.enable_shap:
                self.enable_shap = False
            # currently only Lime is supported for NLP, hardcode in args for now
            self.enable_lime = True
            self.build_nlp_explainers(enable_lime=self.enable_lime, enable_shap=self.enable_shap, data=encoded_data)
        if self.model_input_type == InputType.Image:
            if self.load_image_func is None:
                raise UserValueError('load_image_func must be provided for image explanations')
            if self.enable_shap:
                self.enable_shap = False
            # currently only Lime is supported for CV/ImageInput, hardcode in args for now
            self.enable_lime = True
            self.build_image_explainers(enable_lime=self.enable_lime, enable_shap=self.enable_shap, data=encoded_data)
        # not supported input type
        else:
            pass

    def build_tabular_explainers(self, data: pd.DataFrame, **kwargs) -> None:

        try:
            if kwargs.get('enable_shap'):
                samples = data.sample(n=self.shap_training_samples) if len(data) > self.shap_training_samples else data
                shap_explainer = KernelExplainer(model=self.predict_wrapper, data=samples)
                self.tabular_explainer_shap = shap_explainer

            if kwargs.get('enable_lime'):
                mode = 'regression' if self.model_type == OutputType.Regression else 'classification'
                lime_explainer = LimeTabularExplainer(training_data=data.to_numpy(),
                                                      mode=mode,
                                                      categorical_features=self.categorical_features,
                                                      discretize_continuous=False)
                self.tabular_explainer_lime = lime_explainer
        except Exception:
            raise ArthurUserError(f"Failed to build explainer. "
                                  f"Please ensure your predict() function operates on batches, "
                                  f"taking in a sequence of inputs and returning an output in the format "
                                  f"appropriate for your model output predictions. For more info, consult "
                                  f"https://docs.arthur.ai/user-guide/explainability.html#prediction-function")

    def build_nlp_explainers(self, data: pd.DataFrame, **kwargs) -> None:
        # shap not supported at this time
        if kwargs.get('enable_shap'):
            pass

        if kwargs.get("enable_lime"):
            # we always default to no bow so we get one explanation per word
            lime_explainer = LimeTextExplainer(bow=False, split_expression=self.text_delimiter)
            self.nlp_explainer_lime = lime_explainer

    def build_image_explainers(self, data: pd.DataFrame, **kwargs) -> None:
        if kwargs.get('enable_shap'):
            pass

        if kwargs.get('enable_lime'):
            lime_explainer = LimeImageExplainer()
            self.image_explainer_lime = lime_explainer

    def encode_labels(self, feature_vector: Iterable) -> Iterable:
        """Encodes labels into integers"""
        encoded_fv = []
        for i, val in enumerate(feature_vector):
            cur_label_mapping = self.label_mapping[i]
            if cur_label_mapping is not None:
                # if this is a categorical value that isn't captured by attribute categories
                # update mapping to handle gracefully
                if val not in cur_label_mapping:
                    new_mapped_val = max(cur_label_mapping.values()) + 1
                    cur_label_mapping[val] = new_mapped_val
                    # add value to reverse label mapping as well, so we can decode
                    self.reverse_label_mapping[i][new_mapped_val] = val # type: ignore
                encoded_fv.append(cur_label_mapping[val])
            else:
                encoded_fv.append(val)
        return encoded_fv

    def decode_labels(self, feature_vector: Iterable) -> Iterable:
        """Decodes into strings"""
        decoded_fv = []
        for i, val in enumerate(feature_vector):
            cur_reverse_label_mapping = self.reverse_label_mapping[i]
            if cur_reverse_label_mapping is not None:
                decoded_fv.append(cur_reverse_label_mapping[val])
            else:
                decoded_fv.append(val)
        return decoded_fv

    def predict_wrapper(self, encoded_feature_vectors: Iterable) -> Any:
        """Wraps user supplied predict function. Handles decoding back into original format"""
        decoded_feature_vectors = [self.decode_labels(fv) for fv in encoded_feature_vectors]
        return self.input_predict_func(decoded_feature_vectors)

    def shap_expected_values(self) -> List[float]:
        """Returns expected values from the shap explainer for this model input type"""
        if self.model_input_type == InputType.Tabular:
            shap_explainer = self.tabular_explainer_shap
        # other model types not supported ATM
        else:
            raise UserValueError(f"This explainer does not have {self.SHAP} configured "
                                 f"for the model type: {self.model_input_type}")

        if shap_explainer is None:
            raise UserValueError(f"{self.SHAP} explanations are not enabled for this explainer")

        exp_values = shap_explainer.expected_value

        # always return a list even if one predicted value
        if type(exp_values) == np.ndarray:
            exp_values = exp_values.tolist()
        else:
            exp_values = [exp_values]

        return exp_values

    def explain_tabular_shap(self, raw_feature_vectors: List, **kwargs) -> List[List[List[float]]]:
        """Calculate SHAP values. For one or more inferences.

        :param raw_feature_vectors: An iterable of one or more feature vectors to explain.
        :param kwargs: Additional args to pass to SHAP.

        :return: A List of explanations for each feature vector. A single explanation is a list of lists that contain
            the explanation for each model output by input feature. In this example, the model has 4 features and
            3 output classes.

            .. code-block:: python

                [
                    [-0.011, 0.009, -0.128, -0.161],
                    [-0.009, 0.003, -0.091, -0.065],
                    [0.02, -0.013, 0.220, 0.226]
                ]

        """

        if self.model_input_type != InputType.Tabular:
            raise MethodNotApplicableError(f"Cannot run a {InputType.Tabular} explanation on a {self.model_input_type} model")

        if self.tabular_explainer_shap is None:
            raise MethodNotApplicableError(f"Tabular {self.SHAP} explanations are not enabled for this explainer")

        # swap strings for integers using lookup table
        feature_vectors = np.array([self.encode_labels(fv) for fv in raw_feature_vectors], dtype=np.float64)
        explanations = self.tabular_explainer_shap.shap_values(feature_vectors, **kwargs)

        # format by inference not by class
        if type(explanations) == np.ndarray:
            explanations = [explanations]

        # create 2D array result lists for each inference to organize by inference
        # instead of by class
        results_by_inference = []
        for _ in range(len(raw_feature_vectors)):
            results_by_inference.append([])

        for explanation in explanations:
            explanation_list = explanation.tolist()
            for i, row in enumerate(explanation_list):
                results_by_inference[i].append(row)

        return results_by_inference

    def explain_tabular_lime(self, raw_feature_vectors: List, **kwargs) -> List[List[List[float]]]:
        """
        Calculate lime values. For one or more inferences.

        :param raw_feature_vectors: An iterable of one or more feature vectors to explain.
        :param kwargs: Additional args to pass to LIME.
        :return: A List of explanations for each feature vector. A single explanation is
            a list of lists that contain the explanation for each model output by input feature.
            In this example, the model has 4 features and 3 output classes:

            .. code-block:: python

                [
                    [
                        [-0.011, 0.009, -0.128, -0.161],
                        [-0.009, 0.003, -0.091, -0.065],
                        [0.02, -0.013, 0.220, 0.226]
                    ]
                ]
        """

        if self.model_input_type != InputType.Tabular:
            raise MethodNotApplicableError(f"Cannot run a {InputType.Tabular} explanation on a {self.model_input_type} model")

        if self.tabular_explainer_lime is None:
            raise MethodNotApplicableError(f"Tabular {self.LIME} explanations are not enabled for this explainer")

        # explain all features and classes
        num_features = len(raw_feature_vectors[0])
        if self.model_type == OutputType.Regression:
            # ignored by lime for regression
            labels = [1]
        else:
            # for classification, lime returns
            # results by label in the range 0 to (n-1)
            labels = list(range(self.num_predicted_attributes))

        # swap strings for integers using lookup table
        feature_vectors = np.array([self.encode_labels(fv) for fv in raw_feature_vectors], dtype=np.float64)

        explanations: List[Union[Explanation, EmptyLimeExplanation]] = []
        for feature_vector in feature_vectors:
            # if there are any NaNs in this feature, return all NaNs as the explanation
            if np.any(np.isnan(feature_vector)):
                exp = EmptyLimeExplanation(class_int_labels=labels, num_features=num_features)
            else:
                exp = self.tabular_explainer_lime.explain_instance(feature_vector,
                                                                   self.predict_wrapper,
                                                                   num_features=num_features,
                                                                   labels=labels,
                                                                   **kwargs)
            explanations.append(exp)

        return self.format_lime_results(explanations)

    def explain_tabular(self,
                        algo: str,
                        raw_feature_vectors: List,
                        nsamples: Optional[int] = None,
                        **kwargs) -> List[List[List[float]]]:
        if self.model_input_type != InputType.Tabular:
            raise MethodNotApplicableError(f"Cannot run a {InputType.Tabular} explanation on a {self.model_input_type} model")

        if algo == self.SHAP:
            # shap takes nsamples
            if nsamples is not None:
                kwargs['nsamples'] = nsamples
            return self.explain_tabular_shap(raw_feature_vectors=raw_feature_vectors, **kwargs)

        elif algo == self.LIME:
            # lime takes num_samples
            if nsamples is not None:
                kwargs['num_samples'] = nsamples
            return self.explain_tabular_lime(raw_feature_vectors=raw_feature_vectors, **kwargs)

        # other algos not supported ATM
        else:
            raise UserValueError(f"This explainer does not know how to explain {algo} "
                                 f"for the model type: {self.model_input_type}")

    def explain_nlp_lime(self, raw_feature_vectors: List, **kwargs) -> List[List[List[List[float]]]]:
        if self.model_input_type != InputType.NLP:
            raise MethodNotApplicableError(f"Cannot run a {InputType.NLP} explanation on a {self.model_input_type} model")

        if self.nlp_explainer_lime is None:
            raise MethodNotApplicableError(f"Tabular {self.LIME} explanations are not enabled for this explainer")

        # explain all features and classes
        if self.model_type == OutputType.Regression:
            # ignored by lime for regression
            labels: Iterable = (1,)
        else:
            # for classification, lime returns
            # results by label in the range 0 to (n-1)
            labels = list(range(self.num_predicted_attributes))

        explanations = []
        for feature_vector in raw_feature_vectors:
            if len(feature_vector) != 1:
                raise UserValueError('Cannot use NLP explainer with more than 1 feature')

            # get number of unique words to ensure we get explanation for each word
            num_features = len(re.split(self.text_delimiter, feature_vector[0]))

            # 5000 is default value from Lime
            num_samples = kwargs.get('num_samples') if 'num_samples' in kwargs else 5000
            exp = self.nlp_explainer_lime.explain_instance(feature_vector[0],
                                                           self.input_predict_func,
                                                           labels=labels,
                                                           num_features=num_features,
                                                           num_samples=num_samples)
            explanations.append(exp)

        return self.format_lime_text_results(explanations)

    def explain_nlp(self,
                    algo: str,
                    raw_feature_vectors: List,
                    nsamples: Optional[int] = None,
                    **kwargs) -> List[List[List[List[float]]]]:
        if self.model_input_type != InputType.NLP:
            raise MethodNotApplicableError(f"Cannot run a {InputType.NLP} explanation on a {self.model_input_type} model")

        if algo == self.SHAP:
            raise UserValueError(f"SHAP explainers are not supported at this time for NLP models")
        elif algo == self.LIME:
            # lime takes num_samples
            if nsamples is not None:
                kwargs['num_samples'] = nsamples
            return self.explain_nlp_lime(raw_feature_vectors=raw_feature_vectors, **kwargs)
        else:
            raise UserValueError(f"This explainer does not know how to explain {algo} "
                            f"for the model type: {self.model_input_type}")

    def explain_image_lime(self, image_paths: List, **kwargs) -> List[Dict[str, Any]]:
        if self.model_input_type != InputType.Image:
            raise MethodNotApplicableError(f"Cannot run a {InputType.Image} explanation on a {self.model_input_type} model")

        if self.image_explainer_lime is None:
            raise MethodNotApplicableError(f"Image {self.LIME} explanations are not enabled for this explainer")

        if self.load_image_func is None:
            raise MethodNotApplicableError("load_image function has not been provided")

        # explain all features and classes
        if self.model_type == OutputType.Regression:
            # ignored by lime for regression
            labels: Iterable = (1,)
        else:
            # for classification, lime returns
            # results by label in the range 0 to (n-1)
            labels = list(range(self.num_predicted_attributes))

        explanations = []

        # Add code to default the number of segments for any give image to be 40
        def segmentation_slic_func(original_image):
            DEFAULT_NUM_SEGMENTS = 40
            return seg.slic(original_image, DEFAULT_NUM_SEGMENTS)

        for image_path in image_paths:
            image = self.load_image_func(image_path)
            # 5000 is default value from Lime
            num_samples = kwargs.get('num_samples') if 'num_samples' in kwargs else 5000
            exp = self.image_explainer_lime.explain_instance(image,
                                                             self.input_predict_func,
                                                             labels=labels,
                                                             num_samples=num_samples,
                                                             segmentation_fn=segmentation_slic_func)
            explanations.append(exp)

        return self.format_lime_image_results(explanations)

    def explain_image(self,
                      algo: str,
                      image_paths: List,
                      nsamples: Optional[int] = None,
                      **kwargs) -> List[Dict[str, Any]]:
        if self.model_input_type != InputType.Image:
            raise MethodNotApplicableError(f"Cannot run a {InputType.Image} explanation on a {self.model_input_type} model")

        if algo == self.SHAP:
            raise UserValueError(f"SHAP explainers are not supported at this time for CV/ImageInput models")
        elif algo == self.LIME:
            # lime takes num_samples
            if nsamples is not None:
                kwargs['num_samples'] = nsamples
            return self.explain_image_lime(image_paths=image_paths, **kwargs)
        else:
            raise UserValueError(f"This explainer does not know how to explain {algo} "
                                 f"for the model type: {self.model_input_type}")

    @staticmethod
    def format_lime_results_for_one_class(feature_index_score_pairs: List[Tuple[int, float]]) -> List[float]:
        """
        This method unpacks the results for a single classes's explanation:

            .. code-block:: python

                [
                    (3, -0.1614709564346036),
                    (2, -0.12871035938256958),
                    (0, -0.011225126011010974),
                    (1, 0.009948039509525414)
                ]

        :return: unpacked lime results for a single class

            .. code-block:: python

                [-0.011, 0.009, -0.128, -0.161]

        """
        feature_index_score_pairs_sorted = sorted(feature_index_score_pairs, key=lambda x: x[0])
        return [score[1] for score in feature_index_score_pairs_sorted]

    def format_lime_results(self, results: List[Union[Explanation, EmptyLimeExplanation]]) -> List[List[List[float]]]:
        """
        LIME's .as_map() method returns scores looking like the following:

            .. code-block:: python

                {
                   0:[
                      (3, -0.1614709564346036),
                      (2, -0.12871035938256958),
                      (0, -0.011225126011010974),
                      (1, 0.009948039509525414)
                   ],
                   1:[
                      (2, -0.0914379318506283),
                      (3, -0.06513690080871648),
                      (0, -0.009164069675758498),
                      (1, 0.003622751782692439)
                   ],
                   2:[
                      (3, 0.2266078572433201),
                      (2, 0.22014829123319787),
                      (0, 0.02038919568676947),
                      (1, -0.013570791292217853)
                   ]
                }

        Where the keys are the class labels, and the values are lists of feature index, score tuples for that class.

        This method unpacks the result to a list of lists for EACH explanation:

            .. code-block:: python

                [
                    [-0.011, 0.009, -0.128, -0.161],
                    [-0.009, 0.003, -0.091, -0.065],
                    [0.02, -0.013, 0.220, 0.226]
                ]

        For regression, this will return a list of lists for one class per inference:

            .. code-block:: python

                [
                    [-0.011, 0.009, -0.128, -0.161]
                ]

        """
        formatted_explanations = []
        for expl in results:

            if self.model_type == OutputType.Regression:
                # lime returns regression scores under the class label key 1
                scores: List[Any] = self.format_lime_results_for_one_class(feature_index_score_pairs=expl.as_map()[1])
                # wrap regression scores in a list because there's only 1 model output
                formatted_explanations.append([scores])

            else:
                sorted_class_score_pairs = sorted(list(expl.as_map().items()), key=lambda x: x[0])
                scores = []
                for class_index, feature_index_score_pairs in sorted_class_score_pairs:
                    scores_for_class = self.format_lime_results_for_one_class(
                        feature_index_score_pairs=feature_index_score_pairs)
                    scores.append(scores_for_class)
                formatted_explanations.append(scores)

        return formatted_explanations

    def format_lime_text_results(self, results: List[Explanation]) -> List[List[List[List[float]]]]:
        """Format LIME results for Text models

        LIME's .as_map() method returns scores looking like the following:

        .. code-block:: python

            {0: [('I', -0.1614709564346036), ('a', -0.12871035938256958), ('am', -0.011225126011010974), ('developer', 0.009948039509525414)],
             1: [('am', -0.0914379318506283), ('a', -0.06513690080871648), ('I', -0.009164069675758498), ('developer', 0.003622751782692439)],
             2: [('developer', 0.2266078572433201), ('am', 0.22014829123319787), ('a', 0.02038919568676947), ('I', -0.013570791292217853)]}

        Where the keys are the class labels, and the values are lists of word, score tuples for that class.

        This method unpacks the result to a list of lists for EACH explanation:

        .. code-block:: python

            [
                [['I', -0.011], ['am', 0.009], ['a', -0.128], ['developer', -0.161]],
                [['I', -0.02], ['am', 0.007], ['a', -0.8], ['developer', 0.11]],
                [['I', -0.03], ['am', 0.89], ['a', -0.28], ['developer', -0.16]]
            ]

        For regression, this will return a list of lists for one class per inference:

        .. code-block:: python

            [
                [['I', -0.011], ['am', 0.009], ['a', -0.128], ['developer', -0.161]]
            ]

        :return: formatted results in the following format

            .. code-block:: python

                [
                    [['I', -0.011], ['am', 0.009], ['a', -0.128], ['developer', -0.161]],
                    [['I', -0.02], ['am', 0.007], ['a', -0.8], ['developer', 0.11]],
                    [['I', -0.03], ['am', 0.89], ['a', -0.28], ['developer', -0.16]]
                ]
        """
        formatted_explanations = []
        for exp in results:
            # lime returns explanations with indices, need to map back to token
            raw_map = exp.as_map()
            final_map = {}
            for key in raw_map.keys():
                id_scores = raw_map[key]
                word_scores = exp.domain_mapper.map_exp_ids(id_scores)

                # lime loses ordering, so we have to re-order by position
                word_score_pos = []
                for id_score, word_score in zip(id_scores, word_scores):
                    _id = id_score[0]
                    pos = exp.domain_mapper.indexed_string.string_position(_id)
                    # convert tuples to lists
                    word_score_pos.append((pos, list(word_score)))

                sorted_by_pos = sorted(word_score_pos, key=lambda x:x[0])
                final_map[key] = [x[1] for x in sorted_by_pos]

            if self.model_type == OutputType.Regression:
                # lime stores regression results under class label 1
                formatted_explanations.append([final_map[1]])
            else:
                # if multiple classes, get as a list and sort by class
                # results in [ (class, [scores]), (class, [scores]) ]
                sorted_class_score_pairs = sorted(list(final_map.items()), key=lambda x: x[0])
                # pull out just scores from each tuple
                fmt_exp = [x[1] for x in sorted_class_score_pairs]
                formatted_explanations.append(fmt_exp)

        return formatted_explanations

    @staticmethod
    def format_lime_image_results(results: List[Explanation]) -> List[Dict[str, Any]]:
        """Format LIME results for images

        :returns: formatted output in the following format:

            .. code-block:: python

                {
                    "lime_segment_mask": [
                        [1, 1, 1, 3, 3, 1, 0],
                        [1, 1, 1, 3, 3, 1, 0],
                        [1, 1, 1, 3, 3, 1, 0],
                        [2, 2, 2, 3, 3, 1, 0]
                    ],
                    "lime_region_mapping": {
                        "0": [[0, 0.23], [1, 0.3], [2, 0.001], [3, -0.5]]
                    }
                }
        """
        formatted_explanations = []
        for explanation in results:
            region_scores = explanation.local_exp
            new_region_scores = {}
            for k, v in region_scores.items():
                new_key = str(k.item()) if isinstance(k, np.generic) else str(k)
                new_vals = []
                for val in v:
                    new_val = np.array(val).tolist()
                    new_vals.append(new_val)
                new_region_scores[new_key] = new_vals

            formatted_explanations.append({
                "lime_segment_mask": explanation.segments.tolist(),
                "lime_region_mapping": new_region_scores
            })

        return formatted_explanations



import unittest

if __name__ == '__main__':

    # TODO update test suite to use Nose which automatically discovers tests cases
    testmodules = [
        'arthurai.tests.test_regression_model',
        'arthurai.tests.test_cv_model',
        'arthurai.tests.test_multistage_regression_model',
        'arthurai.tests.test_multistage_multiclass_model',
        'arthurai.tests.test_nlp_model',
    ]

    suite = unittest.TestSuite()

    for t in testmodules:
        try:
            # If the module defines a suite() function, call it to get the suite.
            mod = __import__(t, globals(), locals(), ['suite'])
            suitefn = getattr(mod, 'suite')
            suite.addTest(suitefn())
        except (ImportError, AttributeError):
            # else, just load all the test cases from the module.
            suite.addTest(unittest.defaultTestLoader.loadTestsFromName(t))

    unittest.TextTestRunner().run(suite)






import functools
import inspect


# Base Error
class ArthurError(Exception):
    """
    Base Error for Arthur SDK. This class should not be used directly, Arthur exceptions should inherit from either
    ArthurUserError or ArthurInternalError.
    """
    pass


# Two Secondary Errors which differentiate between exceptions that are the user's fault and Arthur's fault
class ArthurUserError(ArthurError):
    """
    Exception raised due to incorrect user input to the Arthur SDK. Can be used directly but children are preferred.
    """
    pass


class ArthurInternalError(ArthurError):
    """
    Exception raised when user input is correct but an error occurs. Can be used directly but children are preferred.
    """
    pass


# User Exceptions
class MissingParameterError(ArthurUserError):
    """
    Exception raised when parameters supplied to the Arthur SDK are missing.
    """
    pass


class UserValueError(ArthurUserError, ValueError):
    """
    Exception raised when a user supplies an invalid value to the Arthur SDK.
    """
    pass


class UserTypeError(ArthurUserError, TypeError):
    """
    Exception raised when a user supplies an argument of the incorrect type to the Arthur SDK.
    """
    pass


class MethodNotApplicableError(ArthurUserError):
    """
    Exception raised when the method called is not valid for the resource.
    """
    pass


class ResponseClientError(ArthurUserError):
    """
    Exception raised when a 4XX response is received from the API.
    """
    pass


class ForbiddenError(ResponseClientError):
    """
    Exception raised when a 403 Forbidden response is received from the API.
    """
    pass

# Arthur Internal Exceptions
class ExpectedParameterNotFoundError(ArthurInternalError):
    """
    Exception raised when a field or property should be available from Arthur but is unexpectedly missing.
    """
    pass


class InternalValueError(ArthurInternalError, ValueError):
    """

    Exception raised when a value is unexpected.
    """
    pass


class InternalTypeError(ArthurInternalError, TypeError):
    """

    Exception raised when a value is unexpected.
    """
    pass


class ResponseServerError(ArthurInternalError):
    """
    Exception raised when a 5XX response is received from the API.
    """


class ResponseRedirectError(ArthurInternalError):
    """
    Exception raised when a 3XX response is unexpectedly received from the API.
    """


def arthur_excepted(message=None):
    """
    Decorator to wrap user-facing Arthur functions with exception handling that describes to the user whether the error
    is their fault or is our fault and should be reported.
    :param message: an optional message to prefix the error with, should describe the failure e.g. "failed to send
    inferences" or "an error occurred while creating the model."
    :return: the decorator function
    """
    if message is None:
        prefix = ""
    else:
        prefix = message + " because "

    def decorator_arthur_excepted(func):
        @functools.wraps(func)
        def wrapper_arthur_excepted(*args, **kwargs):
            # ensure all required parameters are present: check manually because TypeErrors from internal calls
            #  should not be UserErrors
            try:
                inspect.signature(func).bind(*args, **kwargs)
                success = True
                err_msg = None
            except TypeError as e:
                success = False
                err_msg = str(e)
            if not success:
                raise MissingParameterError(err_msg)

            # call the function
            try:
                return func(*args, **kwargs)
            # if it is a user error simply re-raise
            except ArthurUserError as e:
                raise e
            # otherwise wrap it in a message saying it's not the user's fault
            except ArthurInternalError as e:
                raise ArthurInternalError(prefix + "an internal exception occurred, please report to Arthur") from e
            except Exception as e:
                raise ArthurInternalError(prefix +
                                          "there was an unexpected internal exception, please report to Arthur") from e
        return wrapper_arthur_excepted
    return decorator_arthur_excepted


# TODO: TU-96 can we add custom linting steps to check that base exceptions aren't raised/caught and only arthur
#  exceptions are used?



from typing import Callable

import pandas as pd
import numpy as np
import re
import json

import pytest

from arthurai.explainability.arthur_explainer import ArthurExplainer
from arthurai.common.constants import InputType, OutputType
from arthurai.common.exceptions import ArthurUserError


# FIXTURES

def predict_reg(input_x):
    pass_class = np.array(input_x)[:, 0]
    prediction = 120 / (2 * pass_class)
    return prediction


def predict_class(input_x):
    age = np.array(input_x)[:, 2]
    prediction = np.clip(age / 60, a_min=0., a_max=1.0)
    return np.stack([1 - prediction, prediction], axis=1)


@pytest.fixture
def titanic_df() -> pd.DataFrame:
    return pd.DataFrame({
            'pass_class': [1, 1, 3, 1, 3, 3, 3, 1, 3, 3],
            'sex': ['F', 'F', 'M', 'F', 'M', 'F', 'M', 'M', 'M', 'M'],
            'age': [16.0, 24.0, 19.0, 58.0, 30.0, 22.0, 40.0, 37.0, 65.0, 32.0],
            'fare': [86.5, 49.5042, 8.05, 153.4625, 7.8958, 7.75, 7.8958, 29.7, 7.75, 7.8958],
            'survived': [1, 1, 1, 1, 0, 1, 0, 0, 0, 0]})


# parametrize by output type
@pytest.fixture(params=[OutputType.Multiclass, OutputType.Regression])
def output_type(request) -> OutputType:
    return request.param


# parametrize by lime / shap
@pytest.fixture(params=["lime", "shap"])
def algo(request) -> str:
    return request.param


# parametrize by doing nothing or nullifying input columns as "nan" or None
@pytest.fixture(params=["regular", "nan", "none"])
def X(request, titanic_df: pd.DataFrame, output_type: OutputType) -> pd.DataFrame:
    if output_type == OutputType.Multiclass:
        X = titanic_df.drop(columns="survived")
    elif output_type == OutputType.Regression:
        X = titanic_df.drop(columns=["fare", "survived"])
    else:
        raise ValueError()

    # convert string categorical to int
    X['sex'] = pd.factorize(titanic_df['sex'])[0]

    if request.param == "regular":
        replacement_val = None
    elif request.param == "nan":
        replacement_val = float("nan")
    elif request.param == "none":
        replacement_val = None
    else:
        raise ValueError()

    if replacement_val is not None:
        X['age'].iloc[2] = replacement_val
        X['pass_class'].iloc[3] = replacement_val

    return X


@pytest.fixture
def predict_func(output_type: OutputType) -> Callable:
    if output_type == OutputType.Multiclass:
        return predict_class
    elif output_type == OutputType.Regression:
        return predict_reg
    else:
        raise ValueError()


def significant_feature(output_type: OutputType) -> int:
    # the predict_class function uses only the age input in column 2
    if output_type == OutputType.Multiclass:
        return 2
    # the predict_reg function uses only the pass class input in column 0
    elif output_type == OutputType.Regression:
        return 0
    else:
        raise ValueError()


# TESTS

def test_arthur_explainer_tabular(X: pd.DataFrame, output_type: OutputType, algo: str, predict_func: Callable) -> None:
    # [None, None, ... ] number of columns
    num_cols = len(X.columns)
    label_mapping = [None for _ in range(num_cols)]
    # create explainer
    if algo == "shap":
        enable_shap = True
    elif algo == "lime":
        enable_shap = False
    else:
        raise ValueError()

    # drop rows with NaNs for "training"
    explainer = ArthurExplainer(model_type=output_type,
                                model_input_type=InputType.Tabular,
                                num_predicted_attributes=1,
                                predict_func=predict_func,
                                data=X,
                                enable_shap=enable_shap,
                                label_mapping=label_mapping)
    feature_vector = X.to_numpy().tolist()
    explanations = np.array(explainer.explain_tabular(algo=algo, raw_feature_vectors=feature_vector))

    average_feature_explanations = (
        np.nanmean(  # take average ignoring NaNs
            np.abs(explanations)  # absolute value to look at magnitude of feature importance, not direction
            [:, 0, :],  # pick the first "predicted column" to give us a data_rows X num_features matrix
            axis=0)  # average across all the rows to get average importance of each feature
    )

    # assert most significant feature is expected
    assert np.argmax(average_feature_explanations) == significant_feature(output_type)

    # assert rows with nans produce explanations with all nans
    rows_with_nans = np.nonzero(np.any(np.isnan(explanations), axis=1))[0]
    for i in rows_with_nans:
        assert np.all(np.isnan(explanations[i]))


def test_arthur_explainer_user_error_tabular(X: pd.DataFrame, predict_func: Callable) -> None:
    # [None, None, ... ] number of columns
    num_cols = len(X.columns)
    label_mapping = [None for _ in range(num_cols)]
    # enable_shap is currently always set to True for non NLP data
    enable_shap = True

    def predict_wrong_shape(input_x):
        age = np.array(input_x)[:, 2]
        prediction = np.clip(age / 60, a_min=0., a_max=1.0)
        #incorrect output format due to concatenation along wrong axis should raise an error in caller
        return np.stack([1 - prediction, prediction], axis=0)

    with pytest.raises(ArthurUserError):
        ArthurExplainer(model_type=OutputType.Regression,
                        model_input_type=InputType.Tabular,
                        num_predicted_attributes=1,
                        predict_func=predict_wrong_shape,
                        data=X,
                        enable_shap=enable_shap,
                        label_mapping=label_mapping)


def test_arthur_explainer_nlp_regression() -> None:
    def regression_predict(feature_vecs):
        results = []
        for fv in feature_vecs:
            results.append(np.array([0.2, 0.8]))
        return np.array(results)

    sample_data = pd.DataFrame([
        ['this is a test'],
        ['a second test with more words'],
        ['and a third test']
    ])

    explainer = ArthurExplainer(model_type=OutputType.Regression,
                                model_input_type=InputType.NLP,
                                num_predicted_attributes=1,
                                predict_func=regression_predict,
                                data=sample_data,
                                enable_shap=False,
                                label_mapping=[None])

    raw_feat_vecs = [
        ['first test test'],
        ['and a longer second test']
    ]
    exps = explainer.explain_nlp("lime", raw_feature_vectors=raw_feat_vecs, nsamples=100)

    # confirm we got 2 explanations
    assert len(exps) == 2

    for i, exp in enumerate(exps):
        # ensure single class
        assert len(exp) == 1
        # assert one explanation per unique word
        assert len(exp[0]) == len(re.split(explainer.text_delimiter, raw_feat_vecs[i][0]))


def test_arthur_explainer_nlp_classification() -> None:
    def regression_predict(feature_vecs):
        results = []
        for fv in feature_vecs:
            results.append(np.array([0.2, 0.7, 0.1]))
        return np.array(results)

    sample_data = pd.DataFrame([
        ['this is a test'],
        ['a second test with more words'],
        ['and a third test']
    ])

    explainer = ArthurExplainer(model_type=OutputType.Multiclass,
                                model_input_type=InputType.NLP,
                                num_predicted_attributes=3,
                                predict_func=regression_predict,
                                data=sample_data,
                                enable_shap=False,
                                label_mapping=[None])

    raw_feat_vecs = [
        ['first test test'],
        ['and a longer second test']
    ]
    exps = explainer.explain_nlp("lime", raw_feature_vectors=raw_feat_vecs, nsamples=100)

    # confirm we got 2 explanations
    assert len(exps) == 2

    for i, exp in enumerate(exps):
        # ensure single class
        assert len(exp) == 3
        # assert one explanation per unique word
        assert len(exp[0]) == len(re.split(explainer.text_delimiter, raw_feat_vecs[i][0]))


def test_arthur_explainer_image_classification() -> None:
    def load_image(image_path):
        return np.array([
            [[20, 4, 6], [5, 7, 12], [5, 12, 4]],
            [[20, 4, 106], [5, 7, 112], [5, 12, 104]],
            [[20, 4, 206], [5, 7, 212], [5, 12, 204]],
        ])

    def cv_predict(images):
        results = []
        for image in images:
            results.append(np.array([0.2, 0.7, 0.1]))
        return np.array(results)

    sample_data = pd.DataFrame([
        ['this is a test'],
        ['a second test with more words'],
        ['and a third test']
    ])

    explainer = ArthurExplainer(model_type=OutputType.Multiclass,
                                model_input_type=InputType.Image,
                                num_predicted_attributes=3,
                                predict_func=cv_predict,
                                data=sample_data,
                                enable_shap=False,
                                label_mapping=[None],
                                load_image_func=load_image)

    image_paths = [
        'path/to/image',
        'path/to/another/image'
    ]
    exps = explainer.explain_image("lime", image_paths=image_paths)

    # confirm we got 2 explanations
    assert len(exps) == 2

    for i, exp in enumerate(exps):
        """
        Example Output:
            {
                "lime_segment_mask": [
                    [1, 1, 1, 3, 3, 1, 0],
                    [1, 1, 1, 3, 3, 1, 0],
                    [1, 1, 1, 3, 3, 1, 0],
                    [2, 2, 2, 3, 3, 1, 0]
                ],
                "lime_region_mapping": {
                    "0": [[0, 0.23], [1, 0.3], [2, 0.001], [3, -0.5]]
                }
            }
        """
        # confirm we got an appropriate-sized mask
        assert len(exp["lime_segment_mask"]) == len(load_image(image_paths[i]))

        # sanity check on mask values
        # ideally would actually check schema here
        sample_mask_value = exp["lime_segment_mask"][0][0]
        assert sample_mask_value >= 0
        assert next(iter(exp["lime_region_mapping"].values()))

        # ensure serializable
        assert json.dumps(exp)



import pandas as pd
import pytest

from arthurai import ArthurModel, ArthurAttribute
from arthurai.common.constants import InputType, OutputType, ValueType, Stage
from arthurai.core.attributes import AttributeCategory

from tests.fixtures.mocks import client

TABULAR_MODEL_ID = "tabby-mc-tabface"
TABULAR_MODEL_GROUP_ID = "groupy-mc-groupface"
INT_INPUT = "int_input"
INT_INPUT_ATTR = ArthurAttribute(name=INT_INPUT, value_type=ValueType.Integer, stage=Stage.ModelPipelineInput,
                                 categorical=True, position=0,
                                 categories=[AttributeCategory(value="1"), AttributeCategory(value="2")])
FLOAT_NONINPUT = "float_input"
FLOAT_NONINPUT_ATTR = ArthurAttribute(name=FLOAT_NONINPUT, value_type=ValueType.Float, stage=Stage.NonInputData,
                                      position=0, min_range=0.25, max_range=1.)
GROUND_TRUTH = "ground_truth"
PRED = "pred"
GROUND_TRUTH_ATTR = ArthurAttribute(name=GROUND_TRUTH, value_type=ValueType.Float, stage=Stage.GroundTruth,
                                    position=0, attribute_link=PRED)
PRED_ATTR = ArthurAttribute(name=PRED, value_type=ValueType.Float, stage=Stage.PredictedValue, position=0,
                            attribute_link=GROUND_TRUTH)


MODEL_ATTRIBUTES = [INT_INPUT_ATTR, FLOAT_NONINPUT_ATTR, PRED_ATTR, GROUND_TRUTH_ATTR]

SAMPLE_DATAFRAME = pd.DataFrame({
    INT_INPUT: [1, 2, 1, 2, 1, 2],
    FLOAT_NONINPUT: [1., 0.5, 1., .25, .6, .9],
    PRED: [45.6, 94.2, 61.6, 23., 31.6, 80.3],
    GROUND_TRUTH: [30., 90., 60., 40., 40., 80.]
})

UNINITIALIZED_MODEL_DATA = {
        "partner_model_id": "",
        "input_type": InputType.Tabular,
        "output_type": OutputType.Regression,
        "display_name": "",
        "description": ""
}

MODEL_DATA_WITH_ATTRIBUTES = {
    **UNINITIALIZED_MODEL_DATA,
    "attributes": MODEL_ATTRIBUTES
}


@pytest.fixture
def batch_model(client):
    model_data = {
        **MODEL_DATA_WITH_ATTRIBUTES,
        'is_batch': True
    }
    model = ArthurModel(client=client.client, **model_data)
    model.id = TABULAR_MODEL_ID
    return model


@pytest.fixture
def initial_batch_model(client):
    model_data = {
        **UNINITIALIZED_MODEL_DATA,
        'is_batch': True
    }
    model = ArthurModel(client=client.client, **model_data)
    return model


@pytest.fixture
def unsaved_batch_model(client):
    model_data = {
        **MODEL_DATA_WITH_ATTRIBUTES,
        'is_batch': True
    }
    model = ArthurModel(client=client.client, **model_data)
    return model


@pytest.fixture
def streaming_model(client):
    model_data = {
        **MODEL_DATA_WITH_ATTRIBUTES,
        'is_batch': False
    }
    model = ArthurModel(client=client.client, **model_data)
    model.id = TABULAR_MODEL_ID
    return model



from typing import List, Union, Optional

from dataclasses import dataclass

from arthurai.common.constants import ValueType, Stage
from arthurai.common.exceptions import MissingParameterError, UserValueError
from arthurai.core.base import ArthurBaseJsonDataclass


@dataclass
class AttributeCategory(ArthurBaseJsonDataclass):
    """
    A list of the attribute's categories. An attribute will only have categories if it is marked as categorical.
    """
    value: str
    label: Optional[str] = None

    def __post_init__(self):
        self.value = str(self.value)
        if self.label is not None:
            self.label = str(self.label)


@dataclass
class AttributeBin(ArthurBaseJsonDataclass):
    """
    A list of the attribute's bins. An attribute will only have bins if it is not categorical. The bin start is
     exclusive and the end is inclusive, (continuous_start, continuous_end]. Use Null to represent an open end of a bin.
    """
    continuous_start: Optional[float] = None
    continuous_end: Optional[float] = None


@dataclass
class ArthurAttribute(ArthurBaseJsonDataclass):
    """ArthurAttribute encapsulates data associated with a model attribute

    :param attribute_link: Only applicable for `GroundTruth` or `PredictedValue` staged attributes.
        If stage is equal to `GroundTruth`, this represents the associated `PredictedValue` attribute and vice versa
    :param is_positive_predicted_attribute: Only applicable for `PredictedValue` attributes on a Binary
        Classification model. Should be set to `True` on the positive predicted value attribute.
    :param is_unique: Boolean value used to signal if the values of this attribute are unique.
    :param bins: List of bin cut-offs used to discretize continuous attributes. Use `None` as an open ended value.
        ``[None, 18, 65, None]`` represents the three following bins: ``value < 18, 18 < value < 65, value > 65``
    :param monitor_for_bias: boolean value set to `True` if the attribute should be monitored for bias
    :param max_range: Max value for a continuous attribute
    :param min_range: Min value for a continuous attribute
    :param categorical: Boolean value set to `True` if the attribute has categorical values.
    :param position: The array position of attribute within the stage. Required in the PREDICT_FUNCTION_INPUT stage.
    :param label: Label for attribute. If attribute has an encoded name, a more readable label can be set.
    :param stage: :class:`arthurai.common.constants.Stage` of this attribute in the model pipeline
    :param value_type: :class:`arthurai.common.constants.ValueType` associated with this attributes values
    :param name: Name of the attribute. Attribute names can only contain alpha-numeric characters and underscores
        and cannot start with a number.
    :param categories: [Only for Categorical Attributes] If the attribute is categorical, this will contain the
        attribute's categories. It is required only if the categorical flag is set to true.
    """

    name: str
    value_type: ValueType
    stage: Stage
    id: Optional[str] = None
    label: Optional[str] = None
    position: Optional[int] = None
    categorical: Optional[bool] = False
    min_range: Optional[Union[int, float]] = None
    max_range: Optional[Union[int, float]] = None
    monitor_for_bias: bool = False
    categories: Optional[List[AttributeCategory]] = None
    bins: Optional[List[AttributeBin]] = None
    is_unique: bool = False
    is_positive_predicted_attribute: bool = False
    attribute_link: Optional[str] = None

    def set(self, **kwargs):
        """Set one or many of the available properties of the ArthurAttribute class"""

        # do some basic validation on values to be set
        is_unique = kwargs.get("is_unique", self.categorical)
        categorical = kwargs.get("categorical", self.categorical)

        categories = [AttributeCategory(value=c) for c in kwargs.get("categories")] if "categories" in kwargs.keys() \
            else self.categories

        max_range = kwargs.get('max_range', self.max_range)
        min_range = kwargs.get('min_range', self.min_range)

        if categorical and not is_unique and categories is None:
            raise MissingParameterError("categories is required for non-unique categorical attributes")
        elif not categorical:
            if (min_range and max_range is None) or (min_range is None and max_range):
                raise MissingParameterError("Min and max range must both be set to either numerical values or none")
            elif min_range and max_range and min_range > max_range:
                raise UserValueError(f"Min range must be set to a value less then the max range, received the"
                                f" following (min, max) values: ({min_range}, {max_range})")
            elif categories:
                raise UserValueError("Categories are set for this attribute, they must be set to "
                                "None when categorical is False")

        # Check to make sure that the attributes which the user wants to set are valid
        for attribute_name, attribute_value in kwargs.items():
            if not ArthurAttribute.__annotations__.get(attribute_name):
                raise UserValueError(f"Attribute: '{attribute_name}' can not be set on an ArthurAttribute object")
            self.__setattr__(attribute_name, attribute_value)

            # format categories and bins appropriately
            if attribute_name == "categories":
                self.categories = categories
            if attribute_name == "bins":
                self.bins = [AttributeBin(kwargs.get("bins")[i], kwargs.get("bins")[i + 1])
                             for i in range(len(kwargs.get("bins")) - 1)] if "bins" in kwargs.keys() else self.bins

        return self

    def __str__(self):
        return str(self.to_dict())



import json
import os
from datetime import datetime, timezone

import arthurai.core.util as core_util
import numpy as np
import pandas as pd
import pytest
import responses

from arthurai.common.constants import ValueType, Stage, InputType, OutputType, ImageResponseType, ImageContentType
from arthurai.common.exceptions import UserValueError, MethodNotApplicableError
from arthurai.core.attributes import AttributeCategory, AttributeBin, ArthurAttribute
from arthurai.core.models import ArthurModel, ArthurModelGroup
from http import HTTPStatus
from mock import patch, mock_open, MagicMock
from tests.fixtures.models import regression as reg, multiclass as mclass
from tests.helpers import assert_attributes_equal, mock_post, mock_get, mock_patch
from tests.fixtures.models import regression as reg, multiclass as mclass
from tests.fixtures.models.regression import initial_batch_model as reg_initial_batch_model, \
    unsaved_batch_model as reg_unsaved_batch_model, streaming_model
from tests.fixtures.models.multiclass import initial_biclass_model
from tests.fixtures.mocks import client
from tests.test_request_models import load_json, load_json_string
from tests.test_request_models.fixtures import model_response_json_strings


MODEL_RESPONSE_JSON = [
    load_json(
        os.path.join(os.path.dirname(__file__), 'test_request_models', 'resources', 'model_with_categories.json')),
    load_json(
        os.path.join(os.path.dirname(__file__), 'test_request_models', 'resources', 'model_with_attributes.json'))
]

MODEL_GROUP_ID = "01234567-89ab-cdef-0123-456789abcdef"
MODEL_GROUP_NAME = "Test Model Group"
MODEL_GROUP_DESCRIPTION = "Test Model Group Description"
MODEL_GROUP_ARCHIVED = False
MODEL_GROUP_CREATED_AT = "2022-01-20T23:22:18.185267Z"
MODEL_GROUP_UPDATED_AT = "2022-01-20T23:22:32.914068Z"
MODEL_GROUP_JSON = {
    "id": MODEL_GROUP_ID,
    "name": MODEL_GROUP_NAME,
    "description": MODEL_GROUP_DESCRIPTION,
    "archived": MODEL_GROUP_ARCHIVED,
    "created_at": MODEL_GROUP_CREATED_AT,
    "updated_at": MODEL_GROUP_UPDATED_AT,
    "versions": MODEL_RESPONSE_JSON
}


@responses.activate
def test_send_parquet_file():
    input_dict = {
        "attr1": np.Inf,
        "attr2": "string1",
        "attr3": 3.44,
        "attr4": True,
        "attr5": np.nan,
        "attr6": -np.inf,
        "attr7": "string2",
        "attr8": None,
        "attr9": "",
        "attr10": np.inf,
    }
    expected_dict = {
        "attr1": None,
        "attr2": "string1",
        "attr3": 3.44,
        "attr4": True,
        "attr5": None,
        "attr6": None,
        "attr7": "string2",
        "attr8": None,
        "attr9": "",
        "attr10": None,
    }
    output_dict = ArthurModel._replace_nans_and_infinities_in_dict(input_dict)
    assert expected_dict == output_dict


def test_replace_nans_none_type():
    output_dict = ArthurModel._replace_nans_and_infinities_in_dict(None)
    assert output_dict is None


def test_standardize_pd_obj_for_nans():
    input_df = pd.DataFrame(
        [
            [1, 2, 3., 4, 5, 6, np.inf, np.nan, 1, 1],
            [7, 8, 9., 1, 2, 3, np.nan, -np.inf, 2, 2],
            [4, 5.5, 9., -np.inf, np.nan, np.Inf, np.nan, np.nan, np.nan, np.nan],
        ],
        columns = ["x0", "x1", "x2", "x3", "x4", "x5", "x6", "x7", "x8", "x9"]
    )
    expected_df = pd.DataFrame(
        [
            [1, 2, 3., 4, 5, 6, np.nan, np.nan, 1, 1],
            [7, 8, 9., 1, 2, 3, np.nan, np.nan, 2, 2],
            [4, 5.5, 9., np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
        ],
        columns = ["x0", "x1", "x2", "x3", "x4", "x5", "x6", "x7", "x8", "x9"]
    )
    expected_df.x3 = expected_df.x3.astype(pd.Int64Dtype())
    expected_df.x4 = expected_df.x4.astype(pd.Int64Dtype())
    expected_df.x5 = expected_df.x5.astype(pd.Int64Dtype())

    attributes = {
        "x3": ValueType.Integer,
        "x4": ValueType.Integer,
        "x5": ValueType.Integer,
        "x8": ValueType.Float,
    }

    assert expected_df.equals(
        core_util.standardize_pd_obj(input_df, dropna=False, replacedatetime=False, attributes=attributes)
    )
    with pytest.raises(ValueError):
        assert expected_df.equals(
            core_util.standardize_pd_obj(input_df, dropna=True, replacedatetime=False, attributes=attributes)
        )


def test_standardize_pd_obj_for_series():
    input_series = [
        pd.Series([1, 2, 3]),
        pd.Series([1., 2., 3.]),
        pd.Series([1, 2, np.Inf], name="x1"),
        pd.Series([1, 2, np.Inf], name="x1"),
        pd.Series([1, 2, np.Inf], name="x1"),
        pd.Series([np.nan, -np.inf, np.inf]),
    ]
    input_attributes = [
        {},
        None,
        {"x1": ValueType.Integer},
        {"x1": ValueType.Float},
        {"x2": ValueType.Integer},
        {}
    ]
    expected_undropped_series = [
        pd.Series([1, 2, 3]),
        pd.Series([1., 2., 3.]),
        pd.Series([1, 2, np.nan]).astype(pd.Int64Dtype()),
        pd.Series([1, 2, np.nan]),
        pd.Series([1, 2, np.nan]),
        pd.Series([np.nan, np.nan, np.nan]),
    ]
    expected_dropped_series = [
        pd.Series([1, 2, 3]),
        pd.Series([1., 2., 3.]),
        pd.Series([1, 2]).astype(pd.Int64Dtype()),
        pd.Series([1, 2.]),
        pd.Series([1, 2.]),
        pd.Series([]),
    ]

    for inp, inp_attributes, exp_undropped, exp_dropped in zip(input_series, input_attributes, expected_undropped_series, expected_dropped_series):
        print(inp_attributes)
        assert exp_undropped.equals(
            core_util.standardize_pd_obj(inp, dropna=False, replacedatetime=False, attributes=inp_attributes)
        )
        assert exp_dropped.equals(
            core_util.standardize_pd_obj(inp, dropna=True, replacedatetime=False, attributes=inp_attributes)
        )


def test_standardize_pd_obj_bad_input():
    for inp in ["string", "", 123, [1,2,3], {"key": "value"}]:
        with pytest.raises(TypeError):
            core_util.standardize_pd_obj(inp, dropna=True, replacedatetime=False)


def test_standardize_pd_obj_for_datetimes():
    unaware_timestamp = datetime(2021, 2, 8, 12, 10, 5)
    aware_timestamp = datetime(2021, 2, 8, 12, 10, 5, tzinfo=timezone.utc)
    timestamp_to_expected_output = (
        (unaware_timestamp, ValueError),
        (aware_timestamp, "2021-02-08T12:10:05Z"),
    )

    for timestamp, expected_output in timestamp_to_expected_output:
        if isinstance(expected_output, str):
            actual_output = core_util.standardize_pd_obj(pd.Series([timestamp]), dropna=False, replacedatetime=True)
            assert expected_output == actual_output.values[0]
        else:
            with pytest.raises(expected_output):
                core_util.standardize_pd_obj(pd.Series([timestamp]), dropna=False, replacedatetime=True)


def test_add_attribute_with_categories():
    model = ArthurModel.from_json(model_response_json_strings[2])
    model.add_attribute(
        name="test_categorical_1",
        value_type=ValueType.Integer,
        stage=Stage.ModelPipelineInput,
        categorical=True,
        categories=[1, 2, 3, 4],
    )
    attr = model.get_attribute(name="test_categorical_1")
    assert len(attr.categories) == 4
    assert isinstance(attr.categories[0], AttributeCategory)

    model.add_attribute(
        name="test_categorical_2",
        value_type=ValueType.Integer,
        stage=Stage.ModelPipelineInput,
        categorical=True,
        categories=[AttributeCategory(value="hello")],
    )
    attr = model.get_attribute(name="test_categorical_2")
    assert len(attr.categories) == 1
    assert isinstance(attr.categories[0], AttributeCategory)
    assert attr.categories[0].value == "hello"


def test_add_attribute_with_bins():
    model = ArthurModel.from_json(model_response_json_strings[2])
    model.add_attribute(
        name="test_bins_1",
        value_type=ValueType.Integer,
        stage=Stage.ModelPipelineInput,
        categorical=True,
        bins=[None, 10, 15, 20, None],
    )
    attr = model.get_attribute(name="test_bins_1")
    assert len(attr.bins) == 4
    assert isinstance(attr.bins[0], AttributeBin)
    assert attr.bins[0].continuous_start is None
    assert attr.bins[0].continuous_end == 10

    model.add_attribute(
        name="test_bins_2",
        value_type=ValueType.Integer,
        stage=Stage.ModelPipelineInput,
        categorical=True,
        bins=[
            AttributeBin(continuous_start=None, continuous_end=10),
            AttributeBin(continuous_start=10, continuous_end=15),
            AttributeBin(continuous_start=15, continuous_end=20),
        ],
    )
    attr = model.get_attribute(name="test_bins_2")
    assert len(attr.bins) == 3
    assert isinstance(attr.bins[0], AttributeBin)
    assert attr.bins[0].continuous_start is None
    assert attr.bins[0].continuous_end == 10


def test_add_nlp_attribute():
    model = ArthurModel(
        partner_model_id="test",
        client=None,
        input_type=InputType.NLP,
        output_type=OutputType.Multiclass,
    )
    ref_data = pd.DataFrame(
        {"input_text": ["hello", "hi", "what's up", "yo", "cool"],}
    )
    model.infer_schema(ref_data, stage=Stage.ModelPipelineInput)

    assert (
        model.get_attribute("input_text").value_type == ValueType.Unstructured_Text
    )
    assert model.get_attribute("input_text").categories is None


def test_all_null_column():
    model = ArthurModel(
        partner_model_id="test",
        client=None,
        input_type=InputType.NLP,
        output_type=OutputType.Multiclass,
    )
    df = pd.DataFrame({'col1': [1], 'col2': [float("nan")]})
    model.infer_schema(df, stage=Stage.ModelPipelineInput)

    expected_attributes = [
        ArthurAttribute(name='col1', value_type=ValueType.Integer, stage=Stage.ModelPipelineInput, position=0,
                        categorical=True, categories=[AttributeCategory(value='1', label=None)], is_unique=False),
        ArthurAttribute(name='col2', value_type=ValueType.Float, stage=Stage.ModelPipelineInput, position=1,
                        categorical=False, is_unique=False)]

    assert model.attributes == expected_attributes


def test_nonzero_index_column():
    model = ArthurModel(
        partner_model_id="test",
        client=None,
        input_type=InputType.NLP,
        output_type=OutputType.Multiclass,
    )
    df = pd.DataFrame({'col1': [1], 'col2': [float("nan")]})
    df.index = [5]
    model.infer_schema(df, stage=Stage.ModelPipelineInput)

    expected_attributes = [
        ArthurAttribute(name='col1', value_type=ValueType.Integer, stage=Stage.ModelPipelineInput, position=0,
                        categorical=True, categories=[AttributeCategory(value='1', label=None)], is_unique=False),
        ArthurAttribute(name='col2', value_type=ValueType.Float, stage=Stage.ModelPipelineInput, position=1,
                        categorical=False, is_unique=False)]

    assert model.attributes == expected_attributes


def test_infer_numerical_string():
    model = ArthurModel(
        partner_model_id="test",
        client=None,
        input_type=InputType.Tabular,
        output_type=OutputType.Multiclass,
    )
    df = pd.DataFrame({'col1': ["0", "1", "0", "1", "1", "0"], 'col2': ["0.0", "1.0", "0.0", "1.0", "1.0", "0.0"]})
    model.infer_schema(df, stage=Stage.ModelPipelineInput)

    expected_attributes = [
        ArthurAttribute(name='col1', value_type=ValueType.String, stage=Stage.ModelPipelineInput, position=0,
                        categorical=True, categories=[AttributeCategory(value='0', label=None),
                                                      AttributeCategory(value='1', label=None)], is_unique=False),
        ArthurAttribute(name='col2', value_type=ValueType.String, stage=Stage.ModelPipelineInput, position=1,
                        categorical=True, categories=[AttributeCategory(value='0.0', label=None),
                                                      AttributeCategory(value='1.0', label=None)], is_unique=False)
    ]

    assert_attributes_equal(expected_attributes, model.attributes)


def test_infer_string_categorical_dtype():
    model = ArthurModel(
        partner_model_id="test",
        client=None,
        input_type=InputType.Tabular,
        output_type=OutputType.Multiclass
    )

    df = pd.DataFrame({"col1": list("abca"), "col2": list("bccd")}, dtype="category")
    model.infer_schema(df, stage=Stage.ModelPipelineInput)

    expected_attributes = [
        ArthurAttribute(name='col1', value_type=ValueType.String, stage=Stage.ModelPipelineInput, position=0,
                        categorical=True, categories=[AttributeCategory(value='a', label=None),
                                                      AttributeCategory(value='b', label=None),
                                                      AttributeCategory(value='c', label=None)], is_unique=False),
        ArthurAttribute(name='col2', value_type=ValueType.String, stage=Stage.ModelPipelineInput, position=1,
                        categorical=True, categories=[AttributeCategory(value='d', label=None),
                                                      AttributeCategory(value='b', label=None),
                                                      AttributeCategory(value='c', label=None)], is_unique=False)
    ]
    assert_attributes_equal(expected_attributes, model.attributes)


def test_infer_int_categorical_dtype():
    model = ArthurModel(
        partner_model_id="test",
        client=None,
        input_type=InputType.Tabular,
        output_type=OutputType.Multiclass
    )

    df = pd.DataFrame({"col1": [1, 0, 0, 1], "col2": [1, 2, 1, 0]}, dtype="category")
    model.infer_schema(df, stage=Stage.ModelPipelineInput)

    expected_attributes = [
        ArthurAttribute(name='col1', value_type=ValueType.Integer, stage=Stage.ModelPipelineInput, position=0,
                        categorical=True, categories=[AttributeCategory(value='0', label=None),
                                                      AttributeCategory(value='1', label=None)], is_unique=False),
        ArthurAttribute(name='col2', value_type=ValueType.Integer, stage=Stage.ModelPipelineInput, position=1,
                        categorical=True, categories=[AttributeCategory(value='0', label=None),
                                                      AttributeCategory(value='1', label=None),
                                                      AttributeCategory(value='2', label=None)], is_unique=False)
    ]
    assert_attributes_equal(expected_attributes, model.attributes)


def test_set_categorical_value_labels():
    model = ArthurModel.from_json(model_response_json_strings[2])
    model.add_attribute(
        name="test_categorical_1",
        value_type=ValueType.Integer,
        stage=Stage.ModelPipelineInput,
        categorical=True,
        categories=[1, 2],
    )
    expected_categories = {1: "Male", 2: "Female"}
    model.set_attribute_labels("test_categorical_1", labels=expected_categories)
    cats = model.get_attribute("test_categorical_1").categories
    for category in cats:
        assert int(category.value) in expected_categories
        assert expected_categories[int(category.value)] == category.label


@responses.activate
def test_get_image(client):
    model = ArthurModel(client=client.client,
                        partner_model_id="test",
                        input_type=InputType.Image,
                        output_type=OutputType.Multiclass)
    model.id = '1234567890abcdef'
    image_id = '0a1b2c3d4e5f9687'
    path = '.'
    type = ImageResponseType.RawImage
    file_ext = '.png'
    response_content = 'content'.encode()
    expected_image_file = f"{path}/{type}_{image_id}{file_ext}"

    mock_get(f"/api/v3/models/{model.id}/inferences/images/{image_id}?type={type}", response_content,
             headers={'Content-Type': ImageContentType.Png})
    open_mock = mock_open()
    with patch("arthurai.core.models.open", open_mock, create=True):
        resulting_file = model.get_image(image_id, path, type=type)

    assert resulting_file == expected_image_file
    open_mock.assert_called_with(expected_image_file, 'wb')
    open_mock.return_value.write.assert_called_once_with(response_content)


def test_add_object_detection_output_attributes():
    model = ArthurModel(
        partner_model_id="test",
        client=None,
        input_type=InputType.Image,
        output_type=OutputType.Multiclass,
    )

    # cannot call without proper types
    with pytest.raises(MethodNotApplicableError):
        model.add_object_detection_output_attributes('pred', 'gt', ['class_1', 'class_2'])
    model.input_type = InputType.NLP
    model.output_type = OutputType.ObjectDetection
    with pytest.raises(MethodNotApplicableError):
        model.add_object_detection_output_attributes('pred', 'gt', ['class_1', 'class_2'])

    # cannot call without classes
    model.output_type = OutputType.ObjectDetection
    model.input_type = InputType.Image
    with pytest.raises(UserValueError):
        model.add_object_detection_output_attributes('pred', 'gt', [])

    # names cannot match
    with pytest.raises(UserValueError):
        model.add_object_detection_output_attributes('pred', 'pred', ['class_1', 'class_2'])

    # happy path
    model.add_object_detection_output_attributes('pred', 'gt', ['class_1', 'class_2'])
    assert model.get_attribute('pred').value_type == ValueType.BoundingBox
    assert model.get_attribute('gt').value_type == ValueType.BoundingBox
    assert model.image_class_labels == ['class_1', 'class_2']


def test_build_tabular_model(reg_initial_batch_model):
    pred_gt_map = {reg.PRED: reg.GROUND_TRUTH}
    reg_initial_batch_model.build(data=reg.SAMPLE_DATAFRAME, pred_to_ground_truth_map=pred_gt_map,
                                  non_input_columns=[reg.FLOAT_NONINPUT])

    assert_attributes_equal(reg_initial_batch_model.attributes, reg.MODEL_ATTRIBUTES)
    pd.testing.assert_frame_equal(reg_initial_batch_model.reference_dataframe, reg.SAMPLE_DATAFRAME)


def test_build_tabular_class_model(initial_biclass_model):
    pred_gt_map = {mclass.DOG_PRED: mclass.DOG_GROUND_TRUTH, mclass.CAT_PRED: mclass.CAT_GROUND_TRUTH}
    initial_biclass_model.build(data=mclass.SAMPLE_DATAFRAME, pred_to_ground_truth_map=pred_gt_map,
                                non_input_columns=[mclass.FLOAT_NONINPUT], positive_predicted_attr=mclass.DOG_PRED)

    assert_attributes_equal(initial_biclass_model.attributes, mclass.MODEL_ATTRIBUTES)
    pd.testing.assert_frame_equal(initial_biclass_model.reference_dataframe, mclass.SAMPLE_DATAFRAME)


def test_build_tabular_model_no_refset(reg_unsaved_batch_model):
    pred_gt_map = {reg.PRED: reg.GROUND_TRUTH}
    reg_unsaved_batch_model.build(data=reg.SAMPLE_DATAFRAME, pred_to_ground_truth_map=pred_gt_map,
                                  non_input_columns=[reg.FLOAT_NONINPUT], set_reference_data=False)

    assert_attributes_equal(reg_unsaved_batch_model.attributes, reg.MODEL_ATTRIBUTES)
    assert reg_unsaved_batch_model.reference_dataframe is None


def test_build_nlp():
    model = ArthurModel(
        partner_model_id="test",
        client=None,
        input_type=InputType.NLP,
        output_type=OutputType.Regression,
    )
    ref_data = pd.DataFrame({"input_text": ["hello there", "hi", "what's up", "yo", "cool"],
                             "pred": [90, 80, 50, 31, 26],
                             "gt": [95, 81, 42, 30, 30]})
    model.build(ref_data, pred_to_ground_truth_map={"pred": "gt"})

    assert model.get_attribute("input_text").value_type == ValueType.Unstructured_Text
    assert model.get_attribute("input_text").categories is None
    pd.testing.assert_frame_equal(model.reference_dataframe, ref_data)


def test_build_image():
    model = ArthurModel(
        partner_model_id="test",
        client=None,
        input_type=InputType.Image,
        output_type=OutputType.Regression,
    )
    data = pd.DataFrame({'image': ["path1", "path2"], 'preds': [15.2, 31.6], 'gt': [12, 40]})
    model.build(data, pred_to_ground_truth_map={'preds': 'gt'})

    assert model.get_attribute("image").value_type == ValueType.Image
    assert model.get_attribute("image").categories is None
    pd.testing.assert_frame_equal(model.reference_dataframe, data)


def test_build_object_detection():
    model = ArthurModel(
        partner_model_id="test",
        client=None,
        input_type=InputType.Image,
        output_type=OutputType.ObjectDetection,
    )
    # not actually valid object data but that's okay
    data = pd.DataFrame({'image': ["path1", "path2"], 'preds': [0.8, 0.3], 'gt': [1, 0]})
    with pytest.raises(MethodNotApplicableError):
        model.build(data, pred_to_ground_truth_map={'preds': 'gt'})

def test_build_binary_classifier():
    model = ArthurModel(
        partner_model_id="test",
        client=None,
        input_type=InputType.Tabular,
        output_type=OutputType.Multiclass,
        classifier_threshold = 0.42
    )

    data = pd.DataFrame({'pred0': [0,1,0,1,0,1], 
                        'pred1': [1,0,1,0,1,0],
                        'gt0': [1,1,1,0,0,0],
                        'gt1': [0,0,0,1,1,1]})
    
    model.build(data, pred_to_ground_truth_map = {'pred0': 'gt0', 'pred1': 'gt1'}, positive_predicted_attr= 'pred1')

    assert model.classifier_threshold == 0.42



@responses.activate
def test_save_model_with_refset(reg_unsaved_batch_model):
    reg_unsaved_batch_model.reference_dataframe = reg.SAMPLE_DATAFRAME
    mock_set_ref_data = MagicMock()
    reg_unsaved_batch_model.set_reference_data = mock_set_ref_data
    mock_post("/api/v3/models", response_body={
        'id': reg.TABULAR_MODEL_ID,
        'model_group_id': reg.TABULAR_MODEL_GROUP_ID,
        'version_sequence_num': 1,
    }, status=201)

    reg_unsaved_batch_model.save()

    mock_set_ref_data.assert_called_once_with(data=reg.SAMPLE_DATAFRAME)
    assert len(responses.calls) == 1
    assert os.getenv("ARTHUR_LAST_MODEL_ID") == reg.TABULAR_MODEL_ID
    assert reg_unsaved_batch_model.model_group_id == reg.TABULAR_MODEL_GROUP_ID
    assert reg_unsaved_batch_model.version_sequence_num == 1


@responses.activate
def test_save_model_no_refset(reg_unsaved_batch_model):
    mock_set_ref_data = MagicMock()
    reg_unsaved_batch_model.set_reference_data = mock_set_ref_data
    mock_post("/api/v3/models", response_body={
        'id': reg.TABULAR_MODEL_ID,
        'model_group_id': reg.TABULAR_MODEL_GROUP_ID,
        'version_sequence_num': 1,
    }, status=201)

    reg_unsaved_batch_model.save()

    mock_set_ref_data.assert_not_called()
    assert len(responses.calls) == 1
    assert os.getenv("ARTHUR_LAST_MODEL_ID") == reg.TABULAR_MODEL_ID
    assert reg_unsaved_batch_model.model_group_id == reg.TABULAR_MODEL_GROUP_ID
    assert reg_unsaved_batch_model.version_sequence_num == 1


@responses.activate
def test_save_model_without_model_group_id(reg_unsaved_batch_model):
    # Ensure models can still be saved even when the POST `models/` endpoint does not response with model_group_id
    # For testing backwards compatibility of model versioning SDK changes with pre-model-versioning API versions
    reg_unsaved_batch_model.reference_dataframe = reg.SAMPLE_DATAFRAME
    mock_set_ref_data = MagicMock()
    reg_unsaved_batch_model.set_reference_data = mock_set_ref_data
    mock_post("/api/v3/models", response_body={
        'id': reg.TABULAR_MODEL_ID,
    }, status=201)

    reg_unsaved_batch_model.save()

    mock_set_ref_data.assert_called_once_with(data=reg.SAMPLE_DATAFRAME)
    assert len(responses.calls) == 1
    assert os.getenv("ARTHUR_LAST_MODEL_ID") == reg.TABULAR_MODEL_ID
    assert reg_unsaved_batch_model.model_group_id is None
    assert reg_unsaved_batch_model.version_sequence_num is None


def test_add_attribute_positions(reg_unsaved_batch_model):
    reg_unsaved_batch_model.infer_schema(reg.SAMPLE_DATAFRAME[reg.INT_INPUT], stage=Stage.ModelPipelineInput)
    reg_unsaved_batch_model.infer_schema(reg.SAMPLE_DATAFRAME[reg.FLOAT_NONINPUT], stage=Stage.NonInputData)
    reg_unsaved_batch_model.add_regression_output_attributes({reg.PRED: reg.GROUND_TRUTH}, value_type=ValueType.Integer)
    # there is one actual attribute in each stage so should be two in each with duplicates
    expected_attr_positions = {0, 1}
    for stage in (Stage.ModelPipelineInput, Stage.NonInputData, Stage.PredictedValue, Stage.GroundTruth):
        actual_attr_positions = [attr.position for attr in reg_unsaved_batch_model.attributes if attr.stage == stage]
        assert len(actual_attr_positions) == len(expected_attr_positions)
        assert set(actual_attr_positions) == expected_attr_positions


@responses.activate
def test_find_hotspots(streaming_model):
    model = streaming_model

    response_body = {'data': [{'rules': {},
                               'gt_to_info': {'-1': {'class_f1': 0.8870523415977962,
                                                     'class_precision': 0.8298969072164949,
                                                     'class_recall': 0.9526627218934911,
                                                     'count': 169,
                                                     'pred_to_count': {'-1': 161, '1': 8}},
                                              '1': {'class_f1': 0.9356357927786499,
                                                    'class_precision': 0.9738562091503268,
                                                    'class_recall': 0.9003021148036254,
                                                    'count': 331,
                                                    'pred_to_count': {'-1': 33, '1': 298}}},
                               'precision': 0.9251979650966916,
                               'recall': 0.918,
                               'f1': 0.9192145862795214,
                               'accuracy': 0.918,
                               'impurity': 0.15055200000000002,
                               'n_samples': 500,
                               'feature': 'x2',
                               'cutoff': -3.560836390126483}]}

    mock_get(f"/api/v3/models/{model.id}/enrichments/hotspots/find?metric=f1&threshold=0.999&date=2021-08-04",
             response_body=response_body, status=200)
    model.find_hotspots(metric="f1", threshold=0.999, date="2021-08-04")
    assert len(responses.calls) == 1


@responses.activate
def test_query(streaming_model):
    query_body = {"property": "my_base_prop"}
    response_body = {'query_result': "base query response"}
    mock_post(f"/api/v3/models/{streaming_model.id}/inferences/query",
              response_body=response_body, status=200)

    streaming_model.query(query_body)

    assert len(responses.calls) == 1
    assert responses.calls[0].request.body == json.dumps(query_body)


@responses.activate
def test_query_psi_buckets(streaming_model):
    query_body = {"property": "my_psi_prop"}
    response_body = {'query_result': "psi bucket response"}
    mock_post(f"/api/v3/models/{streaming_model.id}/inferences/query/data_drift_psi_bucket_calculation_table",
              response_body=response_body, status=200)

    streaming_model.query(query_body, query_type="drift_psi_bucket_table")

    # assert request body was query
    assert len(responses.calls) == 1
    assert responses.calls[0].request.body == json.dumps(query_body)


@responses.activate
def test_query_bad_type(streaming_model):
    query_body = {"property": "my_prop"}
    with pytest.raises(UserValueError):
        streaming_model.query(query_body, query_type="foobar")


def test_model_group_from_dict():
    model_group = ArthurModelGroup.from_dict(MODEL_GROUP_JSON)

    assert model_group.id == MODEL_GROUP_ID
    assert model_group.name == MODEL_GROUP_NAME
    assert model_group.description == MODEL_GROUP_DESCRIPTION
    assert model_group.archived is MODEL_GROUP_ARCHIVED  # boolean value
    assert model_group.created_at == MODEL_GROUP_CREATED_AT
    assert model_group.updated_at == MODEL_GROUP_UPDATED_AT


def test_model_group_add_version():
    model = ArthurModel.from_dict(MODEL_RESPONSE_JSON[0])
    model_group = ArthurModelGroup.from_dict(MODEL_GROUP_JSON)
    label = "test_label"
    model_group.add_version(model, label)

    assert model.model_group_id == MODEL_GROUP_ID
    assert model.version_label == label


@responses.activate
def test_model_group_update_name(client):
    model_group = ArthurModelGroup.from_dict(MODEL_GROUP_JSON)
    model_group._update_client(client.client)
    model_group_new_name = "New Name"

    patch_response = MODEL_GROUP_JSON
    patch_response['name'] = model_group_new_name

    mock_patch(f"/api/v3/model_groups/{MODEL_GROUP_JSON['id']}",
               status=HTTPStatus.OK,
               response_body=patch_response)
    model_group.name = model_group_new_name

    assert model_group.name == model_group_new_name


@responses.activate
def test_model_group_update_description(client):
    model_group = ArthurModelGroup.from_dict(MODEL_GROUP_JSON)
    model_group._update_client(client.client)
    model_group_new_description = "New Description"

    patch_response = MODEL_GROUP_JSON
    patch_response['description'] = model_group_new_description


    mock_patch(f"/api/v3/model_groups/{MODEL_GROUP_JSON['id']}",
               status=HTTPStatus.OK,
               response_body=patch_response)
    model_group.description = model_group_new_description

    assert model_group.description == model_group_new_description


def test_get_attribute_wrong_stage():
    model_one = ArthurModel.from_json(model_response_json_strings[1])
    print(model_one.get_attributes(stage=Stage.ModelPipelineInput))
    # cannot call valid attribute name but invalid stage
    with pytest.raises(UserValueError):
        # correct stage is Stage.ModelPipelineInput
        model_one.get_attribute(name='attr_1', stage=Stage.GroundTruth)


def test_get_attribute_names_stage_none():
    model_one = ArthurModel.from_json(model_response_json_strings[1])
    attr_one = model_one.get_attribute_names(stage=None)
    assert len(attr_one) == 2

    model_two = ArthurModel.from_json(model_response_json_strings[2])
    attr_two = model_two.get_attribute_names(stage=None)
    assert len(attr_two) == 1
